using System;
using System.Collections;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/DebugCoreMessage")]
  public class DebugCoreMessage : Entity {
    static float FadeDistance = 512f;

    internal FancyText.Text Text;
    private String dialog;

    private float alpha;

    public DebugCoreMessage(EntityData data, Vector2 offset) : base(data.Position + offset) {
      base.Collidable = false;
      base.Collider = new Hitbox(data.Width, data.Height, 0f, 0f);

      this.Tag |= TagsExt.SubHUD | Tags.PauseUpdate | Tags.FrozenUpdate | Tags.TransitionUpdate;
      this.dialog = data.Attr("dialog", "{no dialog set dumbass}").DialogNlOrKey();
    }

    public override void Added(Scene scene) {
      base.Added(scene);

      Level level = this.Scene as Level;
      var width = Vector2.Distance(
        level.WorldToScreen(this.TopRight),
        level.WorldToScreen(this.TopLeft)
      );

      this.Text = FancyText.Parse(
        text: this.dialog,
        maxLineWidth: (int) width,
        linesPerPage: 40
      );

      if (!(scene as Level).Session.Area.SID.StartsWith("bitsbolts/")) {
        throw new Exception("Dumbass this entity is internal ONLY!");
      }
    }

    public override void Update() {
      base.Update();

      Player player = Scene.Tracker.GetEntity<Player>();
      if (player != null) {
        this.alpha = Ease.CubeInOut(Calc.ClampedMap(Math.Abs(this.X - player.X), 0f, FadeDistance, 1f, 0f));
      }
    }

    public override void Render() {
      Level level = this.Scene as Level;

      Vector2 renderPosition = level.WorldToScreen(this.TopCenter);

      this.Text.DrawJustifyPerLine(
        position: renderPosition,
        justify: new Vector2(0.5f, 0f),
        scale: Vector2.One,
        alpha: this.alpha
      );
    }
  }
}
