using System;
using System.Collections;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Flagstrs = System.Collections.Generic.Dictionary<string, bool>;
using Bitsbolts.Components;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/SaveDataFlagSync")]
  [Tracked]
  public class SaveDataFlagSync : Entity {
    [OnLoad]
    public static void Load() {
      On.Celeste.Session.GetFlag += OnSessionGetFlag;
      On.Celeste.Session.SetFlag += OnSessionSetFlag;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Session.GetFlag -= OnSessionGetFlag;
      On.Celeste.Session.SetFlag -= OnSessionSetFlag;
    }


    internal readonly Flagstrs Flags;
    internal readonly bool LoadFlags = true;
    internal readonly bool SaveFlags = true;
    internal readonly bool UseSIDAsPrefix = true;
    internal String Prefix = "";

    public SaveDataFlagSync(EntityData data, Vector2 offset) : base() {
      var version = data.Int("_version", 1);

      this.Flags = data.Attr("flags", "").ParseFlagstrs(allowInversions: false);
      this.LoadFlags = data.Bool("load", true);
      this.SaveFlags = data.Bool("save", true);

      var prefix = data.Attr("prefix", "");
      var raw = data.Bool("raw", false);
      switch (version) {
        case 1:
          this.UseSIDAsPrefix = !raw;
          break;
        case 2:
          if (raw) {
            this.UseSIDAsPrefix = false;
          } else if (prefix != "") {
            this.Prefix = prefix;
            this.UseSIDAsPrefix = false;
          } else {
            this.UseSIDAsPrefix = true;
          }
          break;
      }

      this.Collidable = false;
      this.Visible = false;
    }

    // do flag load in added to allow entities that may rely on these flags in
    // awake the ability to use them early
    public override void Added(Scene scene) {
      base.Added(scene);

      if (this.UseSIDAsPrefix) {
        this.Prefix = (scene as Level).Session.Area.SID;
      }

      if (this.Prefix != "") {
        this.Prefix += ":";
      }

      foreach (var flag in this.Flags.Keys) {
        var prefixed = this.Prefix + flag;
        if (this.LoadFlags) {
          MainModule.Session.SaveDataFlagsLoad[flag] = prefixed;
        }

        if (this.SaveFlags) {
          MainModule.Session.SaveDataFlagsSave[flag] = prefixed;
        }
      }
    }

    private static bool OnSessionGetFlag(
      On.Celeste.Session.orig_GetFlag orig, Session self, string flag
    ) {
      if (flag != null && MainModule.Session.SaveDataFlagsLoad.ContainsKey(flag)) {
        return SaveData.Instance.HasFlag(MainModule.Session.SaveDataFlagsLoad[flag]);
      }
      return orig(self, flag);
    }

    private static void OnSessionSetFlag(
      On.Celeste.Session.orig_SetFlag orig, Session self, string flag, bool value
    ) {
      if (flag != null && MainModule.Session.SaveDataFlagsSave.ContainsKey(flag)) {
        var prefixed = MainModule.Session.SaveDataFlagsSave[flag];
        if (value) {
          SaveData.Instance.SetFlag(prefixed);
        } else {
          SaveData.Instance.Flags.Remove(prefixed);
        }
      } else {
        orig(self, flag, value);
      }
    }
  }
}
