using System;
using System.Collections;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/SturdyDashBlock")]
  [Tracked]
  [TrackedAs(typeof(DashBlock))]
  public class SturdyDashBlock : DashBlock {
    [OnLoad]
    public static void Load() {
      IL.Celeste.DashBlock.Break_Vector2_Vector2_bool_bool += ILDashBlockBreak;
    }

    [OnUnload]
    public static void Unload() {
      IL.Celeste.DashBlock.Break_Vector2_Vector2_bool_bool -= ILDashBlockBreak;
    }


    public SturdyDashBlock(EntityData data, Vector2 offset, EntityID id) : base(
      position: data.Position + offset,
      tiletype: data.Char("tiletype", '3'),
      width: data.Width,
      height: data.Height,
      blendIn: data.Bool("blendin", false),
      permanent: data.Bool("permanent", true),
      canDash: data.Bool("canDash", true),
      id: id
    ) {
      var OnDashCollide_orig = base.OnDashCollide;
      this.OnDashCollide = (Player player, Vector2 direction) => {
        return this.OnDashed(OnDashCollide_orig, player, direction);
      };
    }

    private static void ILDashBlockBreak(ILContext il) {
      var cursor = new ILCursor(il);

      if (cursor.TryGotoNext(MoveType.After, instr => instr.MatchStfld<Entity>("Collidable"))) {
        Logger.Log("bitsbolts.SturdyDashBlock", $"IL@{cursor.Index}: Adding delegate return in DashBlock.Break before base.Collidable!");

        // if (this.OnBreak()) { return; }
        ILLabel end = cursor.DefineLabel();
        cursor.Emit(OpCodes.Ldarg_0);
        cursor.EmitDelegate<Func<DashBlock, bool>>(SturdyDashBlock.OnBreak);

        cursor.Emit(OpCodes.Brfalse, end);
        cursor.Emit(OpCodes.Ret);
        cursor.MarkLabel(end);
      } else {
        Logger.Log(LogLevel.Error, "bitsbolts.SturdyDashBlock", $"IL@{cursor.Index}: hook failed to find find stfld base.Collidable in DashBlock.Break");
      }
    }

    private static bool OnBreak(DashBlock self) {
      if (self is SturdyDashBlock block) {
        // by keeping collision for a frame we can trick kevins, and can still
        // crush the player
        block.Collidable = true;
        block.Add(new Coroutine(block.BreakRoutine()));
        return true;
      }
      return false;
    }

    private IEnumerator BreakRoutine() {
      base.Collidable = false;
      if (DynamicData.For(this).Get<bool>("permanent")) {
        base.RemoveAndFlagAsGone();
      } else {
        base.RemoveSelf();
      }
      base.DestroyStaticMovers();
      yield break;
    }

    private DashCollisionResults OnDashed(DashCollision orig, Player player, Vector2 direction) {
      var result = orig(player, direction);

      if (player.StateMachine.State == 5 || player.StateMachine.State == 5) {
        if (player.CurrentBooster != null) {
          player.CurrentBooster.PlayerReleased();
          player.CurrentBooster = null;
        }
        player.StateMachine.State = 0;
      }

      return result;
    }
  }
}
