using System;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;


// todo: maybe an entity for screen wrapping? (needs unlocking camera first!)
// todo: if it should move backdrops, make a non-teleporting camera entity
// todo: do something for stylegrounds
//   (maybe if LoopX then Backdrop.Position.X += Delta.X, and viseversa for Y)
//   (idk what to do with non-looping stylegrounds)

namespace Bitsbolts.Entities {
  [Tracked]
  [CustomEntity("Bitsbolts/WorldPortal")]
  public class WorldPortal : Entity {
    [OnLoad]
    public static void Load() {
      On.Celeste.Level.LoadLevel += OnLevelLoadLevel;
      On.Monocle.Tracker.Initialize += TrackerInitialize;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Level.LoadLevel -= OnLevelLoadLevel;
      On.Monocle.Tracker.Initialize -= TrackerInitialize;
    }

    private static void TrackerInitialize(On.Monocle.Tracker.orig_Initialize orig) {
      orig();
      MonocleExt.TrackComponent<Leader>(inherited: true);
    }

    internal Components.WorldCamera Camera;

    public WorldPortal(EntityData data, Vector2 offset) : base(data.Position + offset) {
      base.Visible = false;
      base.Collider = new Hitbox(data.Width, data.Height);

      base.Add(this.Camera = new Components.WorldCamera(
        position: Vector2.Zero,
        target: data.NodesOffset(offset)[0],
        width: data.Width,
        height: data.Height
      ));

      // fixme: we don't use Visible nor Collidable!
      var flags = data.Attr("flags").ParseFlagstrs();
      if (flags.Count > 0) {
        base.Add(new FlagstrListener {
          Flags = flags, TriggerOnAwake = true, OnAll = (enabled) => {
            this.Camera.Active = enabled;
            this.Collidable = enabled;
          }
        });
      }
    }


    public override void Update() {
      base.Update();

      if (!this.Collidable) { return; }
      // foreach (Entity collided in this.CollideAllExcept(Tags.HUD)) {
      foreach (Entity collided in this.CollideAll<Actor>()) {
        if (collided is Player player) {
          SeamlessTeleport(player, this.Camera.Delta);
        } else {
          collided.Position += this.Camera.Delta;
        }
      }
    }


    private static Dictionary<Backdrop, Vector2> BackdropOffset = new Dictionary<Backdrop, Vector2>();
    private void SeamlessTeleport(Player player, Vector2 delta) {
      var level = player.Scene as Level;

      level.Camera.Position += delta;


      // teleport the player...
      player.Position += delta;
      player.Hair.MoveHairBy(delta);

      // fixme: fuck this only needs to be applied on all the teleported actors!!!
      // ...and all leaders (including the one on Player too)
      foreach (Leader leader in this.Scene.Tracker.GetComponents<Leader>()) {
        for (var i = 0; i < leader.PastPoints.Count; i++) {
          leader.PastPoints[i] += delta;
        }

        foreach (var follower in leader.Followers) {
          if (follower != null) {
            follower.Entity.Position += delta;
          }
        }
      }


      // add a temporary offset to looping backdrops that is reset when a
      // non-transition level load occurs
      foreach (Backdrop backdrop in level.Background.Backdrops.Concat(level.Foreground.Backdrops)) {
        var offset = Vector2.Zero;

        if (backdrop.LoopX) {
          backdrop.Position.X += (offset.X = delta.X * backdrop.Scroll.X);
        }

        if (backdrop.LoopY) {
          backdrop.Position.Y += (offset.Y = delta.Y * backdrop.Scroll.Y);
        }

        if (offset != Vector2.Zero) {
          BackdropOffset[backdrop] = offset;
        }
      }
    }

    private static void OnLevelLoadLevel(
      On.Celeste.Level.orig_LoadLevel orig, Level level, Player.IntroTypes intro, bool isFromLoader
    ) {
      // if the backdrop is still in the level on next load, then we can take
      // the liberty to undo our offset changes (since backdrop is only
      // reinitialized by LevelLoader, so our changes are otherwise permanent)
      if (intro != Player.IntroTypes.Transition) {
        foreach (var kvp in BackdropOffset) {
          Backdrop backdrop = kvp.Key;
          Vector2 offset = kvp.Value;
          backdrop.Position -= offset;
        }
        BackdropOffset.Clear();
      }
      orig(level, intro, isFromLoader);
    }
  }
}
