using System;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Entities;


namespace Bitsbolts.Entities {
  [Tracked]
  [CustomEntity("Bitsbolts/WorldPortalDual")]
  public class WorldPortalDual : Entity {
    internal Entities.WorldPortal[] portals = new WorldPortal[2] {null, null};

    public WorldPortalDual(EntityData data, Vector2 offset) : base(data.Position + offset) {
      base.Visible = false;
      base.Active = false;

      var height = data.Height;
      var width = data.Width;

      var portalOffset = Vector2.Zero;
      if (data.Attr("mode", "x") == "x") {
        portalOffset = new Vector2(data.Width / 2, 0);
        width = data.Width / 2;
      } else {
        portalOffset = new Vector2(0, data.Height / 2);
        height = data.Height / 2;
      }


      this.portals[0] = new Entities.WorldPortal(
        new EntityData {
          Width = width,
          Height = height,
          Position = data.Position,
          Nodes = new Vector2[] {data.Nodes[0]},
          Values = new Dictionary<string, object> {
            {"flags", data.Attr("flags", "")},
          },
        },
        offset
      );
      this.portals[1] = new Entities.WorldPortal(
        new EntityData {
          Width = width,
          Height = height,
          Position = data.Nodes[0] + portalOffset,
          Nodes = new Vector2[] {data.Position + portalOffset},
          Values = new Dictionary<string, object> {
            {"flags", data.Attr("flags", "")},
          },
        },
        offset
      );
    }


    public override void Added(Scene scene) {
      base.Added(scene);
      base.RemoveSelf();

      foreach (var portal in portals) {
        scene.Add(portal);
      }
    }
  }
}
