using System;
using System.Reflection;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using Celeste;
using Celeste.Mod;


namespace Bitsbolts {
  internal static class MonocleExt {
    internal static void TrackEntity<T>(bool inherited = false) where T: Entity {
      TrackEntity(typeof(T), inherited);
    }

    internal static void TrackEntity(Type type, bool inherited = false) {
      if (!typeof(Entity).IsAssignableFrom(type)) {
        throw new Exception($"Type {type.FullTypeName()} does not inherit from Monocle.Entity!");
      }

      // already tracked; don't track it again
      if (Tracker.StoredEntityTypes.Contains(type)) { return; }

      Logger.Log("bitsbolts.MonocleExt", $"Adding entity type {type.FullTypeName()} to tracker.");

      if (!type.IsAbstract) {
        if (!Tracker.TrackedEntityTypes.ContainsKey(type)) {
          Tracker.TrackedEntityTypes.Add(type, new List<Type>());
        }
        Tracker.TrackedEntityTypes[type].Add(type);
      }
      Tracker.StoredEntityTypes.Add(type);

      if (inherited) {
        foreach (var subtype in EnumerateSubclasses(type).Enumerable()) {
          if (!subtype.IsAbstract) {
            if (!Tracker.TrackedEntityTypes.ContainsKey(subtype)) {
              Tracker.TrackedEntityTypes.Add(subtype, new List<Type>());
            }
            Tracker.TrackedEntityTypes[subtype].Add(type);
          }
        }
      }
    }

    internal static void TrackComponent<T>(bool inherited = false) where T: Component {
      TrackComponent(typeof(T), inherited);
    }

    internal static void TrackComponent(Type type, bool inherited = false) {
      if (!typeof(Component).IsAssignableFrom(type)) {
        throw new Exception($"Type {type.FullTypeName()} does not inherit from Monocle.Component!");
      }

      // already tracked; don't track it again
      if (Tracker.StoredComponentTypes.Contains(type)) { return; }

      Logger.Log("bitsbolts.MonocleExt", $"Adding component type {type.FullTypeName()} to tracker.");

      if (!type.IsAbstract) {
        if (!Tracker.TrackedComponentTypes.ContainsKey(type)) {
          Tracker.TrackedComponentTypes.Add(type, new List<Type>());
        }
        Tracker.TrackedComponentTypes[type].Add(type);
      }
      Tracker.StoredComponentTypes.Add(type);

      if (inherited) {
        foreach (var subtype in EnumerateSubclasses(type).Enumerable()) {
          if (!subtype.IsAbstract) {
            if (!Tracker.TrackedComponentTypes.ContainsKey(subtype)) {
              Tracker.TrackedComponentTypes.Add(subtype, new List<Type>());
            }
            Tracker.TrackedComponentTypes[subtype].Add(type);
          }
        }
      }
    }

    private static IEnumerator<Type> EnumerateSubclasses(Type type) {
      foreach (Type subtype in Assembly.GetEntryAssembly().GetTypes()) {
        if (type != subtype && type.IsAssignableFrom(subtype)) {
          yield return subtype;
        }
      }
      yield break;
    }
  }
}
