using System;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;


namespace Bitsbolts.Triggers {
  [CustomEntity("Bitsbolts/UnlockCamera")]
  [Tracked]
  public class UnlockCamera : Trigger {
    private static IDetour ILPlayerCameraTargetHook;

    [OnLoad]
    public static void Load() {
      ILPlayerCameraTargetHook = new ILHook(
        typeof(Player).GetProperty("CameraTarget").GetGetMethod(),
        ILPlayerCameraTarget
      );
    }

    [OnUnload]
    public static void Unload() {
      ILPlayerCameraTargetHook?.Dispose();
    }


    public UnlockCamera(EntityData data, Vector2 offset) : base(data, offset) {
      var flags = data.Attr("flags", "").ParseFlagstrs();
      if (flags.Count > 0) { base.Add(new FlagstrTrigger(flags)); }
    }


    private static bool CameraUnlocked = false;
    private static void ILPlayerCameraTarget(ILContext il) {
      var cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(MoveType.After, instr => instr.MatchLdfld<Player>("EnforceLevelBounds"))) {
        Logger.Log(LogLevel.Error, "bitsbolts.Triggers.UnlockCamera", $"IL@{cursor.Index}: Hook failed to find if (this.EnforceLevelBounds) in Player.get_CameraTarget!");
        return;
      }

      // skip EnforceLevelBounds check
      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Func<Player, bool>>(
        (player) => {
          CameraUnlocked = player.CollideCheck<UnlockCamera>();
          return !CameraUnlocked;
        }
      );
      cursor.Emit(OpCodes.And);

      // todo: also hook into CameraLockMode bounds caps here

      if (!cursor.TryGotoNext(MoveType.After, instr => instr.MatchCallvirt<Tracker>("GetEntities"))) {
        Logger.Log(LogLevel.Error, "bitsbolts.Triggers.UnlockCamera", $"IL@{cursor.Index}: Hook failed to find Tracker.GetEntities<Killbox>() in Player.get_CameraTarget!");
        return;
      }

      if (!cursor.TryGotoNext(MoveType.After, instr => instr.MatchLdfld<Entity>("Collidable"))) {
        Logger.Log(LogLevel.Error, "bitsbolts.Triggers.UnlockCamera", $"IL@{cursor.Index}: Hook failed to find if (entity.Collidable) in Player.get_CameraTarget!");
        return;
      }

      cursor.EmitDelegate<Func<bool>>(
        () => { return !CameraUnlocked; }
      );
      cursor.Emit(OpCodes.And);
    }
  }
}
