-- vectors

local function round(num)
  if num >= 0 then
    return math.floor(num + 0.5)
  else
    return math.ceil(num - 0.5)
  end
end


local vector2_mt = {
  __name = "vector2"
}
vector2_mt.__index = vector2_mt

local function vector2(x, y)
  return setmetatable({x = x, y = y}, vector2_mt)
end

function vector2_mt.__add(vec1, vec2)
  return vector2(vec1.x + vec2.x, vec1.y + vec2.y)
end

function vector2_mt.__sub(vec1, vec2)
  return vector2(vec1.x - vec2.x, vec1.y - vec2.y)
end

function vector2_mt.__mul(vec1, vec2)
  if type(vec2) ~= "table" then
    return vector2(vec1.x * vec2, vec1.y * vec2)
  else
    return vector2(vec1.x * vec2.x, vec1.y * vec2.y)
  end
end

function vector2_mt.__div(vec1, vec2)
  if type(vec2) ~= "table" then
    return vector2(vec1.x / vec2, vec1.y / vec2)
  else
    return vector2(vec1.x / vec2.x, vec1.y / vec2.y)
  end
end

-- function vector2_mt.__idiv(vec1, vec2)
--   if type(vec2) ~= "table" then
--     return vector2(vec1.x * vec2, vec1.y * vec2)
--   else
--     return vector2(vec1.x.__idiv(vec2.x), vec1.y // vec2.y)
--   end
-- end

function vector2_mt.__unm(vec1)
  return vector2(-vec1.x, -vec1.y)
end

function vector2_mt.__eq(vec1, vec2)
  return vec1.x == vec2.x and vec1.y == vec2.y
end

function vector2_mt.__len(vec)
  return math.sqrt(vec.x*vec.x + vec.y*vec.y)
end

function vector2_mt.__tostring(vec)
  return "(" .. vec.x .. ", " .. vec.y .. ")"
end


function vector2_mt.floor(vec)
  return vector2(math.floor(vec.x), math.floor(vec.y))
end

function vector2_mt.round(vec)
  return vector2(round(vec.x), round(vec.y))
end

function vector2_mt.idiv(vec1, vec2)
  if type(vec2) ~= "table" then
    return vector2(math.floor(vec1.x / vec2), math.floor(vec1.y / vec2))
  else
    return vector2(math.floor(vec1.x / vec2.x), math.floor(vec1.y / vec2.y))
  end
end

function vector2_mt.normalize(vec)
  return vec / (#vec)
end


return vector2
